import requests
import websocket
import json
import time
from enum import Enum

class GrammarlyFilter(Enum):
    PASSIVE_VOICE = 780623
    NUMERAL_WORD = 788344
    UNUSUAL_PAIR = 1156710
    INTERRUPTER_COMMA = 1443936
    REPETETIVE = 1066724
    REDUNDANT_SPACE_BEFORE_POINT = 1126125
    INCORRECT_TENSE = 1058331
    SINGULAR_NOUN_PLURAL_NUMBER = 863760
    OVERUSED_WORD = 1102577
    CONFUSED_WORD = 643814

class GrammarlyInvalidUserException(Exception):
    pass

class GrammarlyCaptchaException(Exception):
    pass

class GrammarlyDocument:

    def __init__(self, parent):
        self.parent = parent
        self.doc_id = 0
        self.document = None
        self.dialect = "american"
        self.genre = "academic_general"
        self.filtered = []


    def create(self):
        url = "https://dox.grammarly.com/documents"
        headers = self.parent.get_headers()
        headers["referer"] = "https://app.grammarly.com/"
        data = "{}"
        cookies = {"grauth": self.parent.get_token()}
        resp = requests.post(url, headers=headers, data=data, cookies=cookies)
        self.document = resp.json()
        if ("status" in self.document and self.document["status"] == "error" and "info" in self.document and "grauth" in self.document["info"]) or ("error" in self.document and "authorized" in self.document["error"]):
            self.parent.token = None
            self.create()
        else:
            self.doc_id = self.document["id"]


    def get(self, doc_id):
        self.doc_id = doc_id


    def set_title(self, title):
        url = "https://dox.grammarly.com/documents/" + str(self.doc_id)
        headers = self.parent.get_headers()
        headers["referer"] = "https://app.grammarly.com/docs/" + str(self.doc_id)
        data = {"title": title, "id": self.doc_id}
        cookies = {"grauth": self.parent.get_token()}
        resp = requests.put(url, headers=headers, data=str(data), cookies=cookies)
        #return resp.text

    def check_text(self, text):
        ret = []
        try:
            #websocket.enableTrace(True)
            ws = websocket.WebSocket()
            c = "Cookie: grauth=%s; csrf-token=%s;" % (self.parent.get_token(), self.parent.get_csrf_token())
            ws.connect("wss://capi.grammarly.com/freews",  header=[c], origin="https://app.grammarly.com")
            data = json.dumps({"token": None, "sid": 0, "docid": self.doc_id, "client": "web", "protocolVersion": "1.0", "dialect":self.dialect, "action": "start", "id": 0})
            ws.send(data)
            result = json.loads(ws.recv())
            if "error" in result and "authorized" in result["error"]:
                self.parent.get_token(True)
                return self.check_text(text)

            ws.send(json.dumps({"ch":["+0:0:" + text + ":0"],"genre":self.genre,"rev":1,"action":"submit_ot","id":1}))
            result = {"action": "waiting"}
            while(result["action"] != "finished"):
                result = json.loads(ws.recv())
                if "pid" in result:
                    if result["pid"] not in self.filtered:
                        ret.append(result)
                else: ret.append(result)
            ws.close()
            return ret
        except:
            return ret




    def add_filter(self, f):
        self.filtered.append(f.value)
        #print(self.filtered)


    def set_genre(self, genre):
        self.genre = genre


    def set_dialect(self, dialect):
        self.dialect = dialect



class Grammarly:

    def __init__(self, user, password):
        self.user = user
        self.password = password
        self.token = None
        self.headers = {
            "Content-Type":"application/json",
            "Connection":"keep-alive",
            "Origin":"https://app.grammarly.com",
            "Accept":"*/*",
            "Accept-Encoding":"gzip, deflate, br",
            "Accept-Language":"en-US,en;q=0.5",
            "User-Agent":"Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:68.0) Gecko/20100101 Firefox/68.0",
            "x-api-version":"1"
        }


    def set_header(self, header, value):
        self.headers[header] = value


    def get_headers(self):
        return self.headers.copy()


    def create_document(self):
        doc = GrammarlyDocument(self)
        doc.create()
        return doc


    def get_document(self, doc_id):
        doc = GrammarlyDocument(self)
        doc.get(doc_id)
        return doc

        
    def add_user(self, email):
        url = "https://institution.grammarly.com/api/institution/admin/users/add/"
        headers = self.get_headers()
        headers.pop("x-api-version")
        headers["Referer"] = "https://admin-panel.grammarly.com/members"
        headers["Origin"] = "https://admin-panel.grammarly.com"
        data = [{"email": email}]
        cookies = {"grauth": self.get_token(), "csrf-token": self.get_csrf_token()}
        headers["X-Client-Type"] = "adminPanel"
        headers["X-CSRF-Token"] = self.get_csrf_token()
        req = requests.Request("POST", url, headers=headers, data=json.dumps(data), cookies=cookies)
        prepared = req.prepare()
        s = requests.Session()
        resp = s.send(prepared)
        print(resp.text)

    def del_user(self, email):
        url = "https://institution.grammarly.com/api/institution/admin/users/delete/"
        headers = self.get_headers()
        headers.pop("x-api-version")
        headers["Referer"] = "https://admin-panel.grammarly.com/members"
        headers["Origin"] = "https://admin-panel.grammarly.com"
        data = {"emails": [email],"type":"Active","reverse":False}
        cookies = {"grauth": self.get_token(), "csrf-token": self.get_csrf_token()}
        headers["X-Client-Type"] = "adminPanel"
        headers["X-CSRF-Token"] = self.get_csrf_token()
        req = requests.Request("POST", url, headers=headers, data=json.dumps(data), cookies=cookies)
        prepared = req.prepare()
        s = requests.Session()
        resp = s.send(prepared)
        print(resp.text)

    def add_word(self, word):
        url = "https://capi.grammarly.com/api/dict"
        data = "{\"word\": \"%s\"}" % word
        headers = self.get_headers()
        headers["referer"] = "https://app.grammarly.com/profile"
        cookies = {"grauth": self.get_token(), "csrf-token": self.get_csrf_token()}
        resp = requests.post(url, headers=headers, data=data, cookies=cookies)


    def get_words(self):
        url = "https://capi.grammarly.com/api/dict"
        headers = self.get_headers()
        headers["referer"] = "https://app.grammarly.com/profile"
        cookies = {"grauth": self.get_token(), "csrf-token": self.get_csrf_token()}
        resp = requests.get(url, headers=headers, cookies=cookies)
        return resp.json()


    def logout(self):
        url = "https://auth.grammarly.com/v3/logout"
        data = "{}"
        headers = self.get_headers()
        headers["referer"] = "https://app.grammarly.com/"
        cookies = {"grauth": self.get_token(), "csrf-token": self.get_csrf_token()}
        resp = requests.post(url, headers=headers, data=data, cookies=cookies)


    def get_token(self, force = False):
        if force: self.token = None
        if self.token != None: return self.token

        # Get csfr and container id
        url = "https://grammarly.com/"
        resp = requests.get(url)

        self.csrf = resp.cookies["csrf-token"]
        self.set_header('X-CSRF-Token', self.csrf)
        self.set_header('X-Container-Id', resp.cookies['gnar_containerId'])

        url = "https://auth.grammarly.com/v3/api/login"
        data = {
            "email_login": {
                "email" : self.user,
                "password" : self.password
            }
        }
        cookies = {}
        for c in resp.cookies:
            cookies[c.name] = resp.cookies[c.name]
        headers = self.get_headers()
        headers["referer"] = "https://www.grammarly.com/signin"
        resp = requests.post(url, data=str(data), cookies=cookies, headers=headers)
        r = resp.json()

        if 'error' in r and r['error'] == 'SHOW_CAPTCHA':
            raise GrammarlyCaptchaException('CAPTCHA required')

        if "grauth" not in resp.cookies:
            raise GrammarlyInvalidUserException()
        self.token = resp.cookies["grauth"]
        self.csrf = resp.cookies["csrf-token"]
        return self.token

    def get_csrf_token(self):
        return self.csrf

    def set_csrf_token(self, token):
        self.csrf = token

    def set_token(self, token):
        self.token = token


    def check(self, text):
        doc = self.create_document()
        return doc.set_text(text)


    def check_user(self):
        ws = websocket.WebSocket()
        ws.connect("wss://capi.grammarly.com/freews",  header=["Cookie: grauth=" + self.get_token() +"; csrf-token=" + self.get_csrf_token()])
        ws.send(json.dumps({"token": None, "sid": 0, "docid": 0, "client": "web", "protocolVersion": "1.0", "dialect":"american", "action": "start", "id": 0}))
        r = ws.recv()
        print(r)
        result = json.loads(r)
        ws.close()
        if "error" in result and "authorized" in result["error"]:
            self.get_token(True)


