#!/usr/bin/python3
from grammarly import Grammarly, GrammarlyFilter, GrammarlyInvalidUserException, GrammarlyCaptchaException
from latex import LatexParser, LatexPosition
import cookie

from queue import Queue

import sys
import time
import json
import re
import progressbar
import os
import difflib

import concurrent.futures


import settings

NO_TUI = False
NO_GUI = False

try:
    from tui import *
except:
    NO_TUI = True
try:
    from gui import *
except:
    NO_GUI = True

IF_TUI = 0
IF_GUI = 1

interface = IF_GUI

class Correction:
    def __init__(self):
        self.replace_list = {}
        self.replace_list_idx = []
        self.current = 0

    def from_queue(self, parser, q):
        while not q.empty():
            entry = q.get()
            for r in entry[1]:
                if "end" not in r: continue
                off = parser.original_offset(r["begin"], entry[0])
                if off not in self.replace_list: self.replace_list[off] = []
                self.replace_list[off].append(r) #{"title": r["title"], "replace": r["replacements"], "end": parser.original_offset(r["end"], entry[0])})
        self.filter_hints()
        self.parse()


    def to_file(self, fname):
        with open(fname, "w") as outfile:
            json.dump(self.replace_list, outfile)


    def from_file(self, fname):
        try:
            with open(fname) as data_file:
                self.replace_list = json.load(data_file)
            self.filter_hints()
            self.parse()
        except:
            pass


    def filter_hints(self):
        n_replace_list = {}
        for k,v in self.replace_list.items():
            keep = []
            for val in v:
                if "title" not in val: continue
                keep.append(val)
            if len(keep) > 0:
                n_replace_list[int(k)] = keep
        self.replace_list = n_replace_list


    def parse(self):
        n_replace_list = {}
        for k,v in self.replace_list.items():
            n_replace_list[int(k)] = v
            for val in v:
                val["len"] = val["end"] - val["begin"] + 1
        self.replace_list = n_replace_list
        self.replace_list_idx = []
        for key in sorted(self.replace_list.keys()):
                if len(self.replace_list[key]) > 0:
                    self.replace_list_idx.append(key)


    def is_at(self, pos):
        return (pos in self.replace_list and len(self.replace_list[pos]) > 0)


    def start(self, pos):
        for k in self.replace_list:
            if k <= pos and pos <= self.end(k):
                return k
        return -1


    def end(self, pos):
        try:
            res = self.replace_list[pos][0]["len"]
            for m in self.replace_list[pos]:
                res = max(res, m["len"])
            return pos + res
        except:
            return 0


    def from_pos(self, pos):
        for i in range(len(self.replace_list_idx)):
            if self.replace_list_idx[i] <= pos and pos <= self.end(self.replace_list_idx[i]):
                return i
        return -1


    def goto_next(self):
        if self.current < len(self.replace_list_idx): self.current += 1


    def goto_prev(self):
        if self.current > 0: self.current -= 1


    def goto_closest_next(self, pos):
        min_d = 1 << 31
        min_p = len(self.replace_list_idx) - 1

        for i in range(len(self.replace_list_idx)):
            if pos - 1 < self.replace_list_idx[i]:
                if self.replace_list_idx[i] - pos < min_d:
                    min_d = self.replace_list_idx[i] - pos
                    min_p = i

        self.set_current(min_p)


    def goto_closest_prev(self, pos):
        min_d = 1 << 31
        min_p = 0

        for i in range(len(self.replace_list_idx)):
            if self.replace_list_idx[i] < pos - 1:
                if pos - self.replace_list_idx[i] < min_d:
                    min_d = pos - self.replace_list_idx[i]
                    min_p = i
        self.set_current(min_p)


    def current_obj(self):
        if self.current >= 0 and self.current < len(self.replace_list_idx) and self.replace_list_idx[self.current] in self.replace_list:
            return self.replace_list[self.replace_list_idx[self.current]]
        else:
            return None


    def current_id(self):
        return self.current


    def to_pos(self):
        if self.current >= 0 and self.current < len(self.replace_list_idx):
            return self.replace_list_idx[self.current]
        else:
            return 0


    def set_current(self, c):
        if c < 0: c = 0
        if c >= len(self.replace_list_idx): c = len(self.replace_list_idx) - 1
        self.current = c


    def remove(self, index, ui=None):
        if index < 0 or index >= len(self.replace_list_idx): return
        self.replace_list[self.replace_list_idx[index]] = []
        if ui:
            for l in ui.hi_lines:
                for c in l:
                    if c["id"] == index:
                        c["id"] = -1
                    elif c["id"] > index:
                        c["id"] -= 1
        self.parse()


    def ignore(self, index, ui):
        pid = self.replace_list[self.replace_list_idx[index]][0]["pid"]
        to_remove = 0
        cnt = len(self.replace_list_idx)
        for i in range(cnt):
            if to_remove < len(self.replace_list_idx) and len(self.replace_list[self.replace_list_idx[to_remove]]) > 0 and self.replace_list[self.replace_list_idx[to_remove]][0]["pid"] == pid:
                self.remove(to_remove, ui)
            else:
                to_remove += 1
        self.parse()


    def replacement(self, nr):
        if self.current_obj() == None:
            return None
        if nr >= 0 and nr < len(self.current_obj()[0]["replacements"]):
            return self.current_obj()[0]["replacements"][nr]
        else:
            return None


    def replace(self, repl, ui = None):
        if self.current_obj() == None or repl == None:
            return
        delta = len(repl) - self.current_obj()[0]["len"] + 1
        if ui: ui.replace_substr(self.replace_list_idx[self.current], self.current_obj()[0]["len"], repl)

        n_replace_list = {}
        for k,v in self.replace_list.items():
            if k >= self.replace_list_idx[self.current]:
                if k != self.replace_list_idx[self.current]: n_replace_list[k + delta] = v
            else:
                n_replace_list[k] = v
        self.replace_list = n_replace_list
        self.parse()
        if ui: ui.parse(ui.text)


    def correct(self, pos, delta):
        n_replace_list = {}
        for k,v in self.replace_list.items():
            if k >= pos:
                n_replace_list[k + delta] = v
            else:
                n_replace_list[k] = v
        self.replace_list = n_replace_list
        self.parse()




def progress(msg, check = None):
    if interface == IF_TUI:
        print(msg)
    else:
        if check: check.update_status.emit(msg)


def check(data):
    global checked, bar, parent, q, doc
    #try:
    q.put([data[0], doc.check_text(data[1])])
        #q.put([data[0], thread_test(data[1])])
    #except:
        #pass
    checked += 1
    if parent:
        parent.update_progress.emit(100 * checked / to_check)
    else:
        bar.update(checked)


def parse(p=None):
    global checked, to_check, bar, parent, q, doc

    parent = p
    checked = 0
    to_check = 0

    try:
        with open("settings.json") as s:
            settings.form_data = json.load(s)
    except:
        settings.show_settings = True

    try:
        settings.text = open(sys.argv[1]).read()
    except:
        settings.text = ""
        settings.real_check = False

    settings.correction = Correction()

    if parent:
        parent.starting.emit(0)

    if settings.real_check:
        parser = LatexParser(settings.text)
        for k,v in settings.form_data["replace"].items():
            parser.add_replacement(k, lambda kw,arg,v=v: eval(v))

        progress("Parsing file...", parent)
        parser.skip_until(LatexPosition(settings.form_data["Parsing"]))
        parsed = parser.to_paragraphs()
        if len(("".join(parsed)).strip()) < 2:
            parser.skip_until(LatexPosition.START)
            parsed = parser.to_paragraphs()

        try:
            progress("Login to Grammarly...", parent)
            grammarly = Grammarly(settings.form_data["User"], settings.form_data["Pwd"])

            # try to login via browser cookies
            browser_auth = False
            cj = cookie.load()
            if cj:
                for c in cj:
                    if "grammarly" in c.domain:
                        if "grauth" in c.name and len(c.value) > 0:
                            grammarly.set_token(c.value)
                            browser_auth = True
                        if "csrf-token" in c.name:
                            grammarly.set_csrf_token(c.value)

            try:
                grammarly.get_token(not browser_auth)
            except GrammarlyCaptchaException:
                print("Grammarly requires a CAPTCHA :(")
                print("Please enter the values of the cookies from your browser:")
                grammarly.set_token(input("  grauth: ").strip())
                grammarly.set_csrf_token(input("  csrf-token: ").strip())
            doc = grammarly.get_document(settings.form_data["doc_id"])
        except GrammarlyInvalidUserException:
            settings.show_settings = True

        if not settings.show_settings:
            doc.set_genre(settings.genres[settings.form_data["Genre"]]["key"])
            doc.set_dialect("american" if settings.form_data["Dialect"] == 1 else "british")

            for idx,e in enumerate(settings.err_cat):
                if not settings.form_data["cat" + str(idx)]:
                    doc.add_filter(GrammarlyFilter(e["id"]))

            q = Queue()

            to_check = len(parsed)
            bar = progressbar.ProgressBar(max_value=to_check)
            checked = 0

            progress("Checking document...", parent)

            with concurrent.futures.ThreadPoolExecutor(max_workers=settings.form_data["Threads"]) as executor:
                for v in executor.map(check, enumerate(parsed)):
                    pass
                executor.shutdown(wait=True)

            progress("\nDone!", parent)

            settings.correction.from_queue(parser, q)
            settings.correction.to_file("corrections.json")

            settings.form_data["auth"] = grammarly.token
            settings.form_data["doc_id"] = doc.doc_id
            with open("settings.json", "w") as s:
                    json.dump(settings.form_data, s, indent = 2, sort_keys = True)

    else:
        settings.correction.from_file("corrections.json")

    if parent:
        parent.done.emit(0)



def main():
    global interface
    settings.init()

    settings.show_settings = False

    if sys.stdin.isatty():
        interface = IF_TUI
        settings.real_check = True

    if len(sys.argv) == 3 and sys.argv[1] == "--text":
        interface = IF_TUI
        settings.real_check = True
        sys.argv[1] = sys.argv[2]

    if len(sys.argv) >= 2 and sys.argv[1] == "--gui":
        interface = IF_GUI
        settings.real_check = False
        if len(sys.argv) > 2: sys.argv[1] = sys.argv[2]

    if (len(sys.argv) < 2) and interface == IF_TUI:
        print("Usage: %s file.tex" % sys.argv[0])
        sys.exit(0)

    if interface == IF_TUI and NO_TUI:
        print("TUI not supported, are all requirements installed?")
        sys.exit(1)
    if interface == IF_GUI and NO_GUI:
        print("GUI not supported, are all requirements installed?")
        sys.exit(1)



    if interface == IF_TUI:
        parse()
        while True:
            try:
                Screen.wrapper(app)
            except ResizeScreenError:
                if settings.show_settings:
                    if settings.tui != None:
                        settings.tui.screen.clear()
                        settings.tui.screen.refresh()
                    break
            except:
                raise
                sys.exit(0)
        print("Restarting...")
        main()
    else:
        a = QtWidgets.QApplication(sys.argv)
        if settings.show_settings:
            SettingsDialog()
            main()
        else:
            settings.parse = parse
            parse()
            main_app = GUI()
            main_app.show()

            return a.exec_()

if __name__ == "__main__":
    main()
